#include "stdafx.h"
#include <map>
#include "grtpp_module_cpp.h"
#include "cppdbc.h"

#include "grts/structs.db.mgmt.h"

#include <cppdbc.h>

class DbMySQLQueryImpl : public grt::ModuleImplBase
{
public:
  DbMySQLQueryImpl(grt::CPPModuleLoader *loader) 
  : grt::ModuleImplBase(loader),
  _last_error_code(0), _connection_id(0), _resultset_id(0), _tunnel_id(0)
  {
    _mutex = g_mutex_new();
  }

  virtual ~DbMySQLQueryImpl()
  {
    g_mutex_free(_mutex);
  }

  DEFINE_INIT_MODULE("1.0", "MySQL AB", grt::ModuleImplBase,
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::openConnection), 
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::closeConnection),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::lastError),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::lastErrorCode),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::execute),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::executeQuery),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultNumRows),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultNumFields),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldType),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldName),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultNextRow),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldIntValue),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldDoubleValue),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldStringValue),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldIntValueByName),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldDoubleValueByName),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::resultFieldStringValueByName),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::closeResult),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::openTunnel),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::getTunnelPort),
                  DECLARE_MODULE_FUNCTION(DbMySQLQueryImpl::closeTunnel),
                  NULL);

  // returns connection-id or -1 for error
  int openConnection(const db_mgmt_ConnectionRef &info);
  int closeConnection(int conn);

  std::string lastError();
  int lastErrorCode();

  // returns 1/0 for ok, -1 for error
  int execute(int conn, const std::string &query);

  // returns result-id or -1 for error
  int executeQuery(int conn, const std::string &query);

  int resultNumRows(int result);
  int resultNumFields(int result);
  std::string resultFieldType(int result, int field);
  std::string resultFieldName(int result, int field);
  // returns 1 if ok, 0 if no more rows
  int resultNextRow(int result);
  int resultFieldIntValue(int result, int field);
  double resultFieldDoubleValue(int result, int field);
  std::string resultFieldStringValue(int result, int field);
  
  int resultFieldIntValueByName(int result, const std::string &field);
  double resultFieldDoubleValueByName(int result, const std::string &field);
  std::string resultFieldStringValueByName(int result, const std::string &field);
  
  int closeResult(int result);

  
  // open SSH tunnel using the connection info
  // returns tunnel id or 0 if no tunnel needed
  int openTunnel(const db_mgmt_ConnectionRef &info);
  int getTunnelPort(int tunnel);
  int closeTunnel(int tunnel);

private:
  GMutex *_mutex;
  std::map<int, sql::ConnectionWrapper> _connections;
  std::map<int, sql::ResultSet*> _resultsets;
  std::map<int, sql::TunnelConnection*> _tunnels;
  std::string _last_error;
  int _last_error_code;

  int _connection_id;
  int _resultset_id;
  int _tunnel_id;
};



GRT_MODULE_ENTRY_POINT(DbMySQLQueryImpl);

#define CLEAR_ERROR() do { _last_error.clear(); _last_error_code = 0; } while (0)


struct Lock
{
  GMutex *mutex;

  Lock(GMutex *mtx)
  {
    mutex = mtx;
    g_mutex_lock(mutex);
  }

  ~Lock()
  {
    g_mutex_unlock(mutex);
  }
};


int DbMySQLQueryImpl::openConnection(const db_mgmt_ConnectionRef &info)
{
  sql::DriverManager *dm = sql::DriverManager::getDriverManager();
  
  if (!info.is_valid())
    throw std::invalid_argument("connection info is NULL");

  CLEAR_ERROR();
  try
  {
    ++_connection_id;
    sql::ConnectionWrapper conn(dm->getConnection(info));
    Lock lock(_mutex);
    _connections[_connection_id] = conn;
  }
  catch (sql::SQLException &exc)
  {
    _last_error = exc.what();
    _last_error_code = exc.getErrorCode();
    Lock lock(_mutex);
    if (_connections.find(_connection_id) != _connections.end())
      _connections.erase(_connection_id);
    return -1;    
  }
  catch (std::exception &exc)
  {
    _last_error = exc.what();
    Lock lock(_mutex);
    if (_connections.find(_connection_id) != _connections.end())
      _connections.erase(_connection_id);
    return -1;
  }
  
  return _connection_id;
}


int DbMySQLQueryImpl::closeConnection(int conn)
{
  CLEAR_ERROR();

  Lock lock(_mutex);
  if (_connections.find(conn) == _connections.end())
    throw std::invalid_argument("Invalid connection");

  _connections.erase(conn);

  return 0;
}


std::string DbMySQLQueryImpl::lastError()
{
  return _last_error;
}


int DbMySQLQueryImpl::lastErrorCode()
{
  return _last_error_code;
}


int DbMySQLQueryImpl::execute(int conn, const std::string &query)
{
  CLEAR_ERROR();
  
  sql::Connection *con= 0;
  {
    Lock lock(_mutex);
    if (_connections.find(conn) == _connections.end())
      throw std::invalid_argument("Invalid connection");
    con = _connections[conn].get();
  }

  try
  {
    sql::Statement *pstmt = con->createStatement();
    bool ok = pstmt->execute(query);
      
    delete pstmt;
    
    return ok ? 1 : 0;
  }
  catch (sql::SQLException &exc)
  {
    _last_error = exc.what();
    _last_error_code = exc.getErrorCode();
    return -1;
  }
  catch (std::exception &e)
  {
    _last_error = e.what();
    return -1;
  }
  
  return -1;
}


int DbMySQLQueryImpl::executeQuery(int conn, const std::string &query)
{
  CLEAR_ERROR();
  
  sql::Connection *con= 0;
  {
    Lock lock(_mutex);
    if (_connections.find(conn) == _connections.end())
      throw std::invalid_argument("Invalid connection");
    con = _connections[conn].get();
  }

  try
  {
    sql::Statement *pstmt = con->createStatement();
    sql::ResultSet *res = pstmt->executeQuery(query);

    ++_resultset_id;

    Lock lock(_mutex);
    _resultsets[_resultset_id] = res;
    
    delete pstmt;
  }
  catch (sql::SQLException &exc)
  {
    _last_error = exc.what();
    _last_error_code = exc.getErrorCode();
    return -1;
  }
  catch (std::exception &e)
  {
    _last_error = e.what();
    return -1;
  }

  return _resultset_id;
}


int DbMySQLQueryImpl::resultNumRows(int result)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];

  return res->rowsCount();
}


int DbMySQLQueryImpl::resultNumFields(int result)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];
  
  return res->getMetaData()->getColumnCount();
}


std::string DbMySQLQueryImpl::resultFieldType(int result, int field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];
  
  return res->getMetaData()->getColumnTypeName(field);  
}


std::string DbMySQLQueryImpl::resultFieldName(int result, int field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];

  return res->getMetaData()->getColumnName(field);
}


int DbMySQLQueryImpl::resultNextRow(int result)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];  
  return res->next() ? 1 : 0;
}


int DbMySQLQueryImpl::resultFieldIntValue(int result, int field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];

  return res->getInt(field);
}


double DbMySQLQueryImpl::resultFieldDoubleValue(int result, int field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];
  
  return res->getDouble(field);
}


std::string DbMySQLQueryImpl::resultFieldStringValue(int result, int field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");

  sql::ResultSet *res = _resultsets[result];
  return res->getString(field);
}


int DbMySQLQueryImpl::resultFieldIntValueByName(int result, const std::string &field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];

  return res->getInt(field);
}


double DbMySQLQueryImpl::resultFieldDoubleValueByName(int result, const std::string &field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];
  
  return res->getDouble(field);
}


std::string DbMySQLQueryImpl::resultFieldStringValueByName(int result, const std::string &field)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    throw std::invalid_argument("Invalid resultset");
  sql::ResultSet *res = _resultsets[result];
  
  return res->getString(field);
}


int DbMySQLQueryImpl::closeResult(int result)
{
  Lock lock(_mutex);
  CLEAR_ERROR();
  if (_resultsets.find(result) == _resultsets.end())
    return -1;
  sql::ResultSet *res = _resultsets[result];
  delete res;
  _resultsets.erase(result);
  return 0;
}




int DbMySQLQueryImpl::openTunnel(const db_mgmt_ConnectionRef &info)
{
  sql::DriverManager *dm = sql::DriverManager::getDriverManager();
  sql::TunnelConnection *tun = dm->getTunnel(info);
  if (tun)
  {
    _tunnels[++_tunnel_id] = tun;
    return _tunnel_id;
  }
  return 0;
}


int DbMySQLQueryImpl::getTunnelPort(int tunnel)
{
  if (_tunnels.find(tunnel) == _tunnels.end())
    throw std::invalid_argument("Invalid tunnel-id");
  return _tunnels[tunnel]->get_port();
}

int DbMySQLQueryImpl::closeTunnel(int tunnel)
{
  if (_tunnels.find(tunnel) == _tunnels.end())
    throw std::invalid_argument("Invalid tunnel-id");
  delete _tunnels[tunnel];
  _tunnels.erase(tunnel);
  return 0;
}


